import logging
from typing import Callable, Dict, List, Tuple, Union

from mvin import Token, TokenBool, TokenError, TokenErrorTypes, TokenNumber, TokenString


def _coerce_excel_int(value: Union[int, float]) -> Union[int, None]:
    if isinstance(value, bool):
        return None
    if isinstance(value, int):
        return value
    if isinstance(value, float) and value.is_integer():
        return int(value)
    return None


def excel_not(token: Token) -> Token:
    if token is not None and token.type == "OPERAND":
        if token.subtype == "LOGICAL":
            return TokenBool(not token.value)
        elif token.subtype == "NUMBER":
            return TokenBool(not (token.value != 0))
    return TokenError(
        TokenErrorTypes.VALUE,
        f"Expected boolean or number and found {token}",
    )


def excel_iserror(token: Token) -> Token:
    return TokenBool(token is not None and token.subtype == "ERROR")


def excel_search(
    find_text: Union[Token, None],
    within_text: Union[Token, None],
    start_num: Union[Token, None],
) -> Token:
    logging.debug(
        f"excel_lib.excel_search: calling with (find_text= {find_text}, within_text={within_text}, start_num= {start_num} )"
    )
    if within_text is None:
        return TokenError(TokenErrorTypes.VALUE, "Argument within_text cannot be None")
    if within_text.type != "OPERAND":
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Expected value for within_text argument, but found: {within_text}",
        )

    within_text_value = ""
    if within_text.subtype == "ERROR":
        return within_text
    elif within_text.subtype == "TEXT":
        within_text_value = within_text.value
    elif within_text.subtype == "NUMBER":
        within_text_value = str(within_text.value)
    elif within_text.subtype == "LOGICAL":
        within_text_value = "TRUE" if within_text.value else "FALSE"
    else:
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Unsupported value type for within_text argument: {within_text}",
        )

    find_text_value = ""
    if find_text is not None:
        if find_text.subtype == "ERROR":
            return find_text
        elif find_text.subtype == "TEXT":
            find_text_value = find_text.value
        elif find_text.subtype == "NUMBER":
            find_text_value = str(find_text.value)
        elif find_text.subtype == "LOGICAL":
            find_text_value = "TRUE" if find_text.value else "FALSE"
        else:
            return TokenError(
                TokenErrorTypes.VALUE,
                f"Unsupported value type for find_text argument: {find_text}",
            )

    start_num_value = 0
    if start_num is not None:
        if start_num.subtype == "ERROR":
            return start_num
        elif start_num.subtype == "NUMBER":
            coerced_start_num = _coerce_excel_int(start_num.value)
            if coerced_start_num is None:
                return TokenError(
                    TokenErrorTypes.VALUE,
                    f"Expected integer >= 1 for start_num argument, but found: {start_num.value}",
                )
            start_num_value = coerced_start_num - 1

            if start_num_value < 0:
                return TokenError(
                    TokenErrorTypes.VALUE,
                    f"Expected integer >= 1 for start_num argument, but found: {start_num.value}",
                )
        else:
            return TokenError(
                TokenErrorTypes.VALUE,
                f"Expected integer value for start_num argument, but found: {start_num}",
            )

    # Excel SEARCH is case-insensitive.
    found_index = within_text_value.lower().find(find_text_value.lower(), start_num_value)
    if found_index >= 0:
        return TokenNumber(found_index + 1)
    else:
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Text `{find_text_value}` not found in text `{within_text_value} (starting at: {start_num_value + 1})`",
        )


def excel_left(text: Union[Token, None], num_chars: Union[Token, None]) -> Token:
    logging.debug(
        f"excel_lib.excel_left: calling with (text= {text}, num_chars= {num_chars} )"
    )
    if text is None:
        return TokenError(
            TokenErrorTypes.VALUE,
            "Expected value for text argument, but Empty was found",
        )
    if text.type != "OPERAND":
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Expected value for text argument, but found: {text}",
        )

    if num_chars is not None and num_chars.type != "OPERAND":
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Expected positive integer for num_chars argument, but found: {num_chars}",
        )

    text_value = ""
    if text.subtype == "ERROR":
        return text
    elif text.subtype == "TEXT":
        text_value = text.value
    elif text.subtype == "NUMBER":
        text_value = str(text.value)
    elif text.subtype == "LOGICAL":
        text_value = "TRUE" if text.value else "FALSE"
    else:
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Unsupported value type for text argument: {text}",
        )

    num_chars_value = 1
    if num_chars is not None:
        if num_chars.subtype == "ERROR":
            return num_chars
        elif num_chars.subtype == "NUMBER":
            coerced_num_chars = _coerce_excel_int(num_chars.value)
            if coerced_num_chars is None:
                return TokenError(
                    TokenErrorTypes.VALUE,
                    f"Expected positive integer for num_chars argument, but found: {num_chars.value}",
                )
            num_chars_value = coerced_num_chars

            if num_chars_value < 0:
                return TokenError(
                    TokenErrorTypes.VALUE,
                    f"Expected positive integer for num_chars argument, but found: {num_chars.value}",
                )
        else:
            return TokenError(
                TokenErrorTypes.VALUE,
                f"Expected positive integer for num_chars argument, but found: {num_chars}",
            )

    return TokenString(text_value[0:num_chars_value])


def excel_right(text: Union[Token, None], num_chars: Union[Token, None]) -> Token:
    logging.debug(
        f"excel_lib.excel_right: calling with (text= {text}, num_chars= {num_chars} )"
    )
    if text is None:
        return TokenError(
            TokenErrorTypes.VALUE,
            "Expected value for text argument, but Empty was found",
        )
    if text.type != "OPERAND":
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Expected value for text argument, but found: {text}",
        )

    if num_chars is not None and num_chars.type != "OPERAND":
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Expected positive integer for num_chars argument, but found: {num_chars}",
        )

    text_value = ""
    if text.subtype == "ERROR":
        return text
    elif text.subtype == "TEXT":
        text_value = text.value
    elif text.subtype == "NUMBER":
        text_value = str(text.value)
    elif text.subtype == "LOGICAL":
        text_value = "TRUE" if text.value else "FALSE"
    else:
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Unsupported value type for text argument: {text}",
        )

    num_chars_value = 1
    if num_chars is not None:
        if num_chars.subtype == "ERROR":
            return num_chars
        elif num_chars.subtype == "NUMBER":
            coerced_num_chars = _coerce_excel_int(num_chars.value)
            if coerced_num_chars is None:
                return TokenError(
                    TokenErrorTypes.VALUE,
                    f"Expected positive integer for num_chars argument, but found: {num_chars.value}",
                )
            num_chars_value = coerced_num_chars

            if num_chars_value < 0:
                return TokenError(
                    TokenErrorTypes.VALUE,
                    f"Expected positive integer for num_chars argument, but found: {num_chars.value}",
                )
        else:
            return TokenError(
                TokenErrorTypes.VALUE,
                f"Expected positive integer for num_chars argument, but found: {num_chars}",
            )

    if num_chars_value == 0:
        return TokenString("")
    return TokenString(text_value[-num_chars_value:])


def excel_len(text: Union[Token, None]) -> Token:
    logging.debug(f"excel_lib.excel_len: calling with (text= {text} )")
    if text is None:
        return TokenError(
            TokenErrorTypes.VALUE,
            "Expected value for text argument, but Empty was found",
        )
    if text.type != "OPERAND":
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Expected value for text argument, but found: {text}",
        )

    text_value = ""
    if text.subtype == "ERROR":
        return text
    elif text.subtype == "TEXT":
        text_value = text.value
    elif text.subtype == "NUMBER":
        text_value = str(text.value)
    elif text.subtype == "LOGICAL":
        text_value = "TRUE" if text.value else "FALSE"
    else:
        return TokenError(
            TokenErrorTypes.VALUE,
            f"Unsupported value type for text argument: {text}",
        )

    return TokenNumber(len(text_value))


DEFAULT_FUNCTIONS: Dict[str, Tuple[Union[List, None], Callable]] = {
    "NOT(": (
        [
            None
        ],  # default argument list (if None is in the list, that argument is not optional)
        excel_not,
    ),
    "ISERROR(": (
        [
            None
        ],  # default argument list (if None is in the list, that argument is not optional)
        excel_iserror,
    ),
    "SEARCH(": (
        [
            TokenString(""),  # find_text
            None,  # within_text
            TokenNumber(1),  # start_num <- default: 1
        ],  # default argument list (if None is in the list, that argument is not optional)
        excel_search,
    ),
    "LEFT(": (
        [
            None,  # text <- required
            TokenNumber(1),  # num_chars <- Optional, default: 1
        ],  # default argument list (if None is in the list, that argument is not optional)
        excel_left,
    ),
    "RIGHT(": (
        [
            None,  # text <- required
            TokenNumber(1),  # num_chars <- Optional, default: 1
        ],  # default argument list (if None is in the list, that argument is not optional)
        excel_right,
    ),
    "LEN(": (
        [
            None,  # text <- required
        ],  # default argument list (if None is in the list, that argument is not optional)
        excel_len,
    ),
}



import pytest  # required for pytest.raises

from mvin import (
    BaseToken,
    Token,
    TokenBool,
    TokenError,
    TokenErrorTypes,
    TokenNumber,
    TokenOperator,
    TokenString,
)
from mvin.interpreter import get_interpreter


class ManualToken(BaseToken):
    def __init__(self, value, token_type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = token_type
        self._subtype = subtype


class FalseyToken(BaseToken):
    def __init__(self, value, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = "OPERAND"
        self._subtype = subtype

    def __bool__(self):
        return False


def test_abstractToken():
    with pytest.raises(TypeError):
        _ = Token()  # pyright: ignore


def test_get_interpreter_ok():
    tokens = [TokenBool(True)]
    run = get_interpreter(tokens)
    assert run is not None


def test_single_bool_ok():
    tokens = [TokenBool(True)]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({})


def test_single_text_ok():
    tokens = [TokenString("hi")]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == "hi"


def test_single_text_neq_1():
    tokens = [TokenString("hi")]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) != 1


def test_single_number():
    tokens = [TokenNumber(1)]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == 1


def test_error_ok():
    error_token = TokenError(TokenErrorTypes.GETTING_DATA, "As expected")
    tokens = [error_token]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({})
    assert result == "#GETTING_DATA"
    assert error_token.message == "As expected"


def test_range_missing_args():
    tokens = [ManualToken("F7", "OPERAND", "RANGE")]
    run = get_interpreter(tokens)
    assert run is not None
    with pytest.raises(KeyError):
        _ = run({})


def test_range_proper_value_transfer():
    tokens = [ManualToken("F7", "OPERAND", "RANGE")]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({"F7": TokenString("done")})
    assert result == "done"


def test_range_falsey_input_token_is_accepted():
    tokens = [ManualToken("F7", "OPERAND", "RANGE")]
    run = get_interpreter(tokens)
    assert run is not None
    result = run({"F7": FalseyToken(0, "NUMBER")})
    assert result == 0


def test_registered_ops_are_copied_defensively():
    def add_v1(a, b):
        return TokenNumber(10)

    def add_v2(a, b):
        return TokenNumber(99)

    custom_ops = {"+": add_v1}
    run = get_interpreter(
        [TokenNumber(1), TokenOperator("+"), TokenNumber(2)],
        registered_ops=custom_ops,
    )
    assert run is not None

    custom_ops["+"] = add_v2
    assert run({}) == 10


def test_function_map_is_copied_defensively():
    def f_v1(value):
        return TokenNumber(value.value + 1)

    def f_v2(value):
        return TokenNumber(value.value + 100)

    custom_functions = {"INC(": ([None], f_v1)}
    tokens = [
        ManualToken("INC(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    run = get_interpreter(tokens, proposed_functions=custom_functions)
    assert run is not None

    custom_functions["INC("] = ([None], f_v2)
    assert run({}) == 2


def test_function_can_return_falsey_token():
    def f_falsey(value):
        return FalseyToken(value.value - 1, "NUMBER")

    custom_functions = {"DEC(": ([None], f_falsey)}
    tokens = [
        ManualToken("DEC(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    run = get_interpreter(tokens, proposed_functions=custom_functions)
    assert run is not None
    assert run({}) == 0


def test_none_as_token():
    tokens = [None]
    with pytest.raises(SyntaxError):
        get_interpreter(tokens)  # pyright: ignore


def test_none_as_list():
    run_as_none = get_interpreter(None)  # pyright: ignore
    assert run_as_none is None



import pytest  # required for pytest.raises

from mvin import BaseToken, TokenBool, TokenNumber, TokenString
from mvin.interpreter import get_interpreter


class ManualToken(BaseToken):
    def __init__(self, value, token_type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = token_type
        self._subtype = subtype


def test_incomplete_call():
    tokens = [ManualToken("SEARCH(", "FUNC", "OPEN")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Unmatched `(` (missing closing parenthesis)."


def test_non_existing_func():
    tokens = [ManualToken("ROBIN(", "FUNC", "OPEN"), ManualToken(")", "FUNC", "CLOSE")]
    with pytest.raises(SyntaxError):
        get_interpreter(tokens)


def test_excel_not_func_incomplete_call():
    tokens = [ManualToken("NOT(", "FUNC", "OPEN"), ManualToken(")", "FUNC", "CLOSE")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Missing required argument at 0 for function `NOT(`"


def test_double_close():
    tokens = [
        ManualToken("(", "PAREN", "OPEN"),
        ManualToken(")", "PAREN", "CLOSE"),
        ManualToken(")", "PAREN", "CLOSE"),
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Unexpected `)` at position 2 (too many closing parentheses)."


def test_operator_after_open():
    tokens = [
        ManualToken("(", "PAREN", "OPEN"),
        ManualToken("+", "OPERATOR-INFIX", ""),
        ManualToken(")", "PAREN", "CLOSE"),
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Unexpected operator `Token<v:+ t:OPERATOR-INFIX s: >` at position 1."


def test_missing_operator_before_par():
    tokens = [TokenNumber(2), ManualToken("(", "PAREN", "OPEN")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Missing operator before '(' at position 1."


def test_handle_missing_last_argument():
    tokens = [
        ManualToken("NOT(", "FUNC", "OPEN"),
        TokenBool(True),
        ManualToken(",", "SEP", "ARG"),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Function `NOT(` expects 1 arguments but got 2."


def test_handle_missing_argument():
    tokens = [
        ManualToken("NOT(", "FUNC", "OPEN"),
        ManualToken(",", "SEP", "ARG"),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Function `NOT(` expects 1 arguments but got 2."


def test_handle_WSPACE():
    tokens = [ManualToken("", "WHITE-SPACE", ""), TokenBool(True)]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({})


def test_raise_unrecognized_token():
    tokens = [ManualToken("XX", "OPERATOR", "")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Unrecognized token `Token<v:XX t:OPERATOR s: >` at position 0."


def test_raise_unmatched_par():
    tokens = [ManualToken("(", "PAREN", "OPEN")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Unmatched `(` (missing closing parenthesis)."


def test_raise_missing_operator_before_open_par():
    tokens = [TokenNumber(2), ManualToken("(", "PAREN", "OPEN")]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Missing operator before '(' at position 1."


def test_too_many_args():
    tokens = [
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenString("hi"),
        ManualToken(",", "SEP", "ARG"),
        TokenString("hi world"),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(1),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(1),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert str(exc_info.value) == "Function `SEARCH(` expects 3 arguments but got 4."


def test_separator_outside_function_is_rejected():
    tokens = [
        ManualToken("(", "PAREN", "OPEN"),
        TokenNumber(1),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(2),
        ManualToken(")", "PAREN", "CLOSE"),
    ]
    with pytest.raises(SyntaxError) as exc_info:
        get_interpreter(tokens)
    assert (
        str(exc_info.value)
        == "Unexpected separator `Token<v:, t:SEP s:ARG >` at position 2."
    )



from mvin import TokenEmpty, TokenParen


def test_token_paren_and_empty():
    open_paren = TokenParen("OPEN")
    close_paren = TokenParen("CLOSE")
    empty = TokenEmpty()

    assert open_paren.value == "("
    assert open_paren.type == "OPERATOR-INFIX"
    assert open_paren.subtype == "OPEN"

    assert close_paren.value == ")"
    assert close_paren.type == "OPERATOR-INFIX"
    assert close_paren.subtype == "CLOSE"

    assert empty.value is None
    assert empty.type == "OPERAND"
    assert empty.subtype == "EMPTY"



import pytest  # required for pytest.raises

from mvin import BaseToken, TokenBool, TokenNumber, TokenParen, TokenString
from mvin.interpreter import get_interpreter


class ManualToken(BaseToken):
    def __init__(self, value, token_type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = token_type
        self._subtype = subtype


def test_excel_not_func_ok_result():
    tokens = [
        ManualToken("NOT(", "FUNC", "OPEN"),
        TokenBool(True),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert not f({})


def test_multiple_operations():
    tokens = [
        ManualToken("(", "PAREN", "OPEN"),
        TokenNumber(2),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(0),
        ManualToken(")", "PAREN", "CLOSE"),
        ManualToken("*", "OPERATOR-INFIX", ""),
        ManualToken("(", "PAREN", "OPEN"),
        TokenNumber(10),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(2),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(3),
        ManualToken(")", "PAREN", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == 36


def test_multiple_operations_with_token_paren():
    tokens = [
        TokenParen("OPEN"),
        TokenNumber(2),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(0),
        TokenParen("CLOSE"),
        ManualToken("*", "OPERATOR-INFIX", ""),
        TokenParen("OPEN"),
        TokenNumber(10),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(2),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(3),
        TokenParen("CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == 36


def test_search_with_leading_whitespace_keeps_separator_indexing():
    tokens = [
        ManualToken("", "WHITE-SPACE", ""),
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenString("a"),
        ManualToken(",", "SEP", "ARG"),
        TokenString("abc"),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == 1


def test_complex_search_call_with_warning():
    tokens = [
        TokenNumber(0),
        ManualToken("+", "OPERATOR-INFIX", ""),
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(1),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(5),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(2),
        ManualToken(")", "PAREN", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == 1


def test_complex_search_call():
    tokens = [
        TokenNumber(2),
        ManualToken("+", "OPERATOR-INFIX", ""),
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(1),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(5),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(2),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == 3


def test_complex_search_call_returns_not_found():
    tokens = [
        TokenNumber(0),
        ManualToken("+", "OPERATOR-INFIX", ""),
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenNumber(5),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(2),
        ManualToken(",", "SEP", "ARG"),
        TokenNumber(1),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(1),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == "#VALUE!"


def test_incomplete_definition():
    tokens = [TokenNumber(0), TokenNumber(1)]
    f = get_interpreter(tokens)
    assert f is not None
    with pytest.raises(ValueError) as exc_info:
        f({})
    assert str(exc_info.value) == "Formula evaluation error: too many values remaining."


def test_double_open_par():
    tokens = [TokenNumber(0), ManualToken("+", "OPERATOR-INFIX", "")]
    f = get_interpreter(tokens)
    assert f is not None
    with pytest.raises(ValueError) as exc_info:
        f({})
    assert str(exc_info.value) == "Not enough values for operation '+'."


def test_required_missing_args():
    tokens = [
        ManualToken("SEARCH(", "FUNC", "OPEN"),
        TokenNumber(1),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(1),
        ManualToken(",", "SEP", "ARG"),
        ManualToken(",", "SEP", "ARG"),
        ManualToken(")", "FUNC", "CLOSE"),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    with pytest.raises(ValueError) as exc_info:
        f({})
    assert str(exc_info.value) == "Missing required argument at 1 for function `SEARCH(`"


def test_unary_minus_simple():
    tokens = [ManualToken("-", "OPERATOR-PREFIX", ""), TokenNumber(2)]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == -2


def test_unary_minus_after_infix():
    tokens = [
        TokenNumber(1),
        ManualToken("+", "OPERATOR-INFIX", ""),
        ManualToken("-", "OPERATOR-PREFIX", ""),
        TokenNumber(2),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == -1


def test_unary_minus_has_lower_precedence_than_exponent():
    tokens = [
        ManualToken("-", "OPERATOR-PREFIX", ""),
        TokenNumber(2),
        ManualToken("^", "OPERATOR-INFIX", ""),
        TokenNumber(2),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == -4


def test_unary_minus_can_be_exponent_operand():
    tokens = [
        TokenNumber(2),
        ManualToken("^", "OPERATOR-INFIX", ""),
        ManualToken("-", "OPERATOR-PREFIX", ""),
        TokenNumber(2),
    ]
    f = get_interpreter(tokens)
    assert f is not None
    assert f({}) == 0.25



from mvin import BaseToken, TokenBool, TokenError, TokenErrorTypes, TokenNumber, TokenString
from mvin.functions.excel_lib import excel_search


class ManualToken(BaseToken):
    def __init__(self, value, token_type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = token_type
        self._subtype = subtype


def test_search_within_text_required():
    result = excel_search(TokenString("a"), None, TokenNumber(1))  # pyright: ignore
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value


def test_search_within_text_invalid_type():
    result = excel_search(
        TokenString("a"),
        ManualToken("x", "FUNC", "OPEN"),
        TokenNumber(1),
    )
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value


def test_search_within_text_error_passthrough():
    within_text = TokenError(TokenErrorTypes.REF, "input error")
    result = excel_search(TokenString("a"), within_text, TokenNumber(1))
    assert result is within_text


def test_search_within_text_logical_conversion():
    result = excel_search(TokenString("RU"), TokenBool(True), TokenNumber(1))
    assert isinstance(result, TokenNumber)
    assert result.value == 2


def test_search_find_text_logical_conversion():
    result = excel_search(TokenBool(False), TokenString("XXFALSEYY"), TokenNumber(1))
    assert isinstance(result, TokenNumber)
    assert result.value == 3


def test_search_find_text_error_passthrough():
    find_text = TokenError(TokenErrorTypes.VALUE, "bad find text")
    result = excel_search(find_text, TokenString("abc"), TokenNumber(1))
    assert result is find_text


def test_search_find_text_unsupported_subtype():
    result = excel_search(
        ManualToken("x", "OPERAND", "RANGE"),
        TokenString("abc"),
        TokenNumber(1),
    )
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value


def test_search_start_num_error_passthrough():
    start_num = TokenError(TokenErrorTypes.VALUE, "bad start")
    result = excel_search(TokenString("a"), TokenString("abc"), start_num)
    assert result is start_num


def test_search_start_num_rejects_non_integer_number():
    result = excel_search(TokenString("a"), TokenString("abc"), TokenNumber(1.5))
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value



from hypothesis import given, settings
from hypothesis import strategies as st

from mvin import BaseToken, TokenNumber
from mvin.interpreter import get_interpreter


class ManualToken(BaseToken):
    def __init__(self, value, token_type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = token_type
        self._subtype = subtype


@settings(max_examples=75, deadline=None)
@given(
    a=st.integers(min_value=-10_000, max_value=10_000),
    b=st.integers(min_value=-10_000, max_value=10_000),
    c=st.integers(min_value=-10_000, max_value=10_000),
)
def test_property_operator_precedence_matches_python_for_add_mul(a, b, c):
    tokens = [
        TokenNumber(a),
        ManualToken("+", "OPERATOR-INFIX", ""),
        TokenNumber(b),
        ManualToken("*", "OPERATOR-INFIX", ""),
        TokenNumber(c),
    ]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == a + b * c


@settings(max_examples=75, deadline=None)
@given(value=st.integers(min_value=-10_000, max_value=10_000))
def test_property_unary_minus_matches_python(value):
    tokens = [ManualToken("-", "OPERATOR-PREFIX", ""), TokenNumber(value)]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == -value


@settings(max_examples=75, deadline=None)
@given(
    left=st.integers(min_value=-10_000, max_value=10_000),
    right=st.integers(min_value=-10_000, max_value=10_000),
)
def test_property_unary_minus_after_infix(left, right):
    tokens = [
        TokenNumber(left),
        ManualToken("+", "OPERATOR-INFIX", ""),
        ManualToken("-", "OPERATOR-PREFIX", ""),
        TokenNumber(right),
    ]
    run = get_interpreter(tokens)
    assert run is not None
    assert run({}) == left + (-right)



from mvin import BaseToken, TokenBool, TokenError, TokenErrorTypes, TokenNumber, TokenString
from mvin.functions.excel_lib import excel_left, excel_len, excel_right


class ManualToken(BaseToken):
    def __init__(self, value, token_type, subtype) -> None:
        super().__init__()
        self._value = value
        self._type = token_type
        self._subtype = subtype


def test_left_basic_and_type_conversions():
    assert excel_left(TokenString("hello"), TokenNumber(2)).value == "he"
    assert excel_left(TokenNumber(12345), TokenNumber(3)).value == "123"
    assert excel_left(TokenBool(True), TokenNumber(4)).value == "TRUE"


def test_left_default_num_chars():
    result = excel_left(TokenString("hello"), None)
    assert isinstance(result, TokenString)
    assert result.value == "h"


def test_left_invalid_input_paths():
    assert excel_left(None, TokenNumber(1)).value == TokenErrorTypes.VALUE.value  # pyright: ignore
    assert (
        excel_left(ManualToken("x", "FUNC", "OPEN"), TokenNumber(1)).value
        == TokenErrorTypes.VALUE.value
    )
    assert (
        excel_left(TokenString("hello"), ManualToken("x", "FUNC", "OPEN")).value
        == TokenErrorTypes.VALUE.value
    )
    assert (
        excel_left(TokenString("hello"), TokenString("2")).value
        == TokenErrorTypes.VALUE.value
    )
    assert (
        excel_left(TokenString("hello"), TokenNumber(-1)).value
        == TokenErrorTypes.VALUE.value
    )
    assert (
        excel_left(TokenString("hello"), TokenNumber(1.5)).value
        == TokenErrorTypes.VALUE.value
    )


def test_left_error_passthroughs():
    text_error = TokenError(TokenErrorTypes.REF, "bad text")
    chars_error = TokenError(TokenErrorTypes.VALUE, "bad chars")
    assert excel_left(text_error, TokenNumber(1)) is text_error
    assert excel_left(TokenString("hello"), chars_error) is chars_error


def test_left_unsupported_text_subtype():
    result = excel_left(ManualToken("x", "OPERAND", "RANGE"), TokenNumber(1))
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value


def test_right_basic_and_type_conversions():
    assert excel_right(TokenString("hello"), TokenNumber(2)).value == "lo"
    assert excel_right(TokenNumber(12345), TokenNumber(3)).value == "345"
    assert excel_right(TokenBool(False), TokenNumber(5)).value == "FALSE"


def test_right_default_num_chars():
    result = excel_right(TokenString("hello"), None)
    assert isinstance(result, TokenString)
    assert result.value == "o"


def test_right_invalid_input_paths():
    assert excel_right(None, TokenNumber(1)).value == TokenErrorTypes.VALUE.value  # pyright: ignore
    assert (
        excel_right(ManualToken("x", "FUNC", "OPEN"), TokenNumber(1)).value
        == TokenErrorTypes.VALUE.value
    )
    assert (
        excel_right(TokenString("hello"), ManualToken("x", "FUNC", "OPEN")).value
        == TokenErrorTypes.VALUE.value
    )
    assert (
        excel_right(TokenString("hello"), TokenString("2")).value
        == TokenErrorTypes.VALUE.value
    )
    assert (
        excel_right(TokenString("hello"), TokenNumber(-1)).value
        == TokenErrorTypes.VALUE.value
    )
    assert (
        excel_right(TokenString("hello"), TokenNumber(1.5)).value
        == TokenErrorTypes.VALUE.value
    )


def test_right_zero_num_chars_returns_empty_string():
    result = excel_right(TokenString("hello"), TokenNumber(0))
    assert isinstance(result, TokenString)
    assert result.value == ""


def test_right_error_passthroughs():
    text_error = TokenError(TokenErrorTypes.REF, "bad text")
    chars_error = TokenError(TokenErrorTypes.VALUE, "bad chars")
    assert excel_right(text_error, TokenNumber(1)) is text_error
    assert excel_right(TokenString("hello"), chars_error) is chars_error


def test_right_unsupported_text_subtype():
    result = excel_right(ManualToken("x", "OPERAND", "RANGE"), TokenNumber(1))
    assert isinstance(result, TokenError)
    assert result.value == TokenErrorTypes.VALUE.value


def test_len_basic_and_type_conversions():
    assert excel_len(TokenString("hello")).value == 5
    assert excel_len(TokenNumber(12345)).value == 5
    assert excel_len(TokenBool(True)).value == 4


def test_len_invalid_input_paths():
    assert excel_len(None).value == TokenErrorTypes.VALUE.value  # pyright: ignore
    assert excel_len(ManualToken("x", "FUNC", "OPEN")).value == TokenErrorTypes.VALUE.value
    assert excel_len(ManualToken("x", "OPERAND", "RANGE")).value == TokenErrorTypes.VALUE.value


def test_len_error_passthrough():
    text_error = TokenError(TokenErrorTypes.REF, "bad text")
    assert excel_len(text_error) is text_error
